@@ -116,7 +116,7 @@ gpdf_control_enable_ui (GPdfControl *control, gboolean sensitive)
 {
 	BonoboUIComponent *uic;
 	GPdfView *gpdf_view;
-	gboolean sens_go_back, sens_go_fwd, sens_zoom;
+	gboolean sens_print, sens_go_back, sens_go_fwd, sens_zoom;
 
 	if (!control->priv->activated)
 		return;
@@ -125,8 +125,12 @@ gpdf_control_enable_ui (GPdfControl *control, gboolean sensitive)
 
 	bonobo_ui_component_freeze (uic, NULL);
 
-	ui_component_set_sensitive (uic, "/commands/FilePrint", sensitive);
+	sens_print = (sensitive
+		      && control->priv->pdf_doc
+		      && control->priv->pdf_doc->okToPrint ());
+	ui_component_set_sensitive (uic, "/commands/FilePrint", sens_print);
 	ui_component_set_sensitive (uic, "/commands/FileProperties", sensitive);
+	ui_component_set_sensitive (uic, "/commands/FileSaveAs", sensitive);
 
 	gpdf_view = GPDF_VIEW (control->priv->gpdf_view);
 	ui_component_set_sensitive (
@@ -160,10 +164,8 @@ gpdf_print_job_prepare (GPdfPrintJob *pj)
 
 	config = gnome_print_job_get_config (pj->gnome_print_job);
 
-	width = height = -1.0;
-
 	/* width and height are returned in PS points :-) */
-	/* gnome_print_config_get_page_size (config, &width, &height); */
+	gnome_print_config_get_page_size (config, &width, &height);
 
 	globalParams->setPSPaperWidth ((int)width);
 	globalParams->setPSPaperHeight ((int)height);
@@ -485,6 +487,60 @@ gpdf_control_properties_dialog_response_cb (GtkDialog *dialog, gint response_id,
 		gtk_widget_destroy (GTK_WIDGET (dialog));
 }
 
+/* Returns a filename in on-disk encoding (suitable for Xpdf code). */
+static gchar *
+gpdf_control_get_filename (BonoboControl *control)
+{
+	GPdfControlPrivate *priv; 
+	GtkWidget *fileselection;
+	GtkWidget *ok_button;
+	GtkWidget *cancel_button;
+	gchar *pathname = NULL; 
+
+	fileselection = gtk_file_selection_new (_("Save a Copy"));
+	gtk_container_set_border_width (GTK_CONTAINER (fileselection), 10);
+	gtk_window_set_position (GTK_WINDOW (fileselection), GTK_WIN_POS_MOUSE);
+	gtk_window_set_modal (GTK_WINDOW (fileselection), TRUE);
+	
+	ok_button = GTK_FILE_SELECTION (fileselection)->ok_button;
+	gtk_widget_show (ok_button);
+	GTK_WIDGET_SET_FLAGS (ok_button, GTK_CAN_DEFAULT);
+	
+	cancel_button = GTK_FILE_SELECTION (fileselection)->cancel_button;
+	gtk_widget_show (cancel_button);
+	GTK_WIDGET_SET_FLAGS (cancel_button, GTK_CAN_DEFAULT);
+
+	bonobo_control_set_transient_for(control,
+					 GTK_WINDOW (fileselection),
+					 NULL);
+	gtk_widget_show (fileselection);
+
+	if (gtk_dialog_run(GTK_DIALOG (fileselection)) == GTK_RESPONSE_OK)
+		pathname = (gchar *)g_strdup (
+			gtk_file_selection_get_filename(
+				GTK_FILE_SELECTION (fileselection)));
+
+	gtk_widget_destroy (fileselection); 
+	
+	return pathname; 
+}
+
+static void
+verb_SaveAs_cb (BonoboUIComponent *uic, gpointer user_data,
+		const char *cname)
+{
+	gchar *filename;
+	
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
+
+	filename = gpdf_control_get_filename (BONOBO_CONTROL (user_data));
+	if (filename != NULL)
+	  gpdf_view_save_as (
+	      GPDF_VIEW (GPDF_CONTROL (user_data)->priv->gpdf_view),
+	      filename);
+	g_free (filename);
+}
+
 static void
 verb_FileProperties_cb (BonoboUIComponent *uic, gpointer user_data,
 			const char *cname)
@@ -862,6 +918,7 @@ gc_set_zoom_items_visibility (GPdfControl *control, gboolean visible)
 
 
 static BonoboUIVerb gc_verbs [] = {
+	BONOBO_UI_VERB ("FileSaveAs", 	  verb_SaveAs_cb),
 	BONOBO_UI_VERB ("FilePrint",      verb_FilePrint_cb),
 	BONOBO_UI_VERB ("FileProperties", verb_FileProperties_cb),
 
@@ -1217,10 +1274,14 @@ gpdf_control_set_property (GObject *object, guint param_id,
 static void
 gpdf_control_setup_persist (GPdfControl *control)
 {
+	gpdf_persist_stream_set_control (control->priv->persist_stream,
+					 BONOBO_CONTROL (control));	
 	g_signal_connect (G_OBJECT (control->priv->persist_stream),
 			  "set_pdf",
 			  G_CALLBACK (persist_stream_set_pdf_cb),
 			  control);
+	gpdf_persist_file_set_control (control->priv->persist_file,
+				       BONOBO_CONTROL (control));
 	g_signal_connect (G_OBJECT (control->priv->persist_file),
 			  "loading_finished",
 			  G_CALLBACK (persist_file_loading_finished_cb),
