@@ -29,6 +29,7 @@
 #include "gpdf-g-switch.h"
 #  include <gdk/gdkkeysyms.h>
 #  include <libgnome/gnome-macros.h>
+#  include <libgnome/gnome-i18n.h>
 #  include <libgnomeprint/gnome-print.h>
 #  include <libgnomeprintui/gnome-print-preview.h>
 #  include <ggv-document.h>
@@ -60,8 +61,8 @@ struct _GPdfViewPrivate {
 	gint drag_anchor_x, drag_anchor_y;
 	gint drag_offset_x, drag_offset_y;
 
-	G_List *forward_history; 
-	G_List *backward_history; 
+	G_List *forward_history;
+	G_List *backward_history;
 };
 
 #define PARENT_TYPE GNOME_TYPE_CANVAS
@@ -103,7 +104,7 @@ gpdf_view_page_background_hide (GPdfView *gpdf_view)
 }
 
 static void
-gpdf_view_page_background_resize (GPdfView *gpdf_view, 
+gpdf_view_page_background_resize (GPdfView *gpdf_view,
 				  gdouble width, gdouble height)
 {
 	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
@@ -117,7 +118,7 @@ gpdf_view_page_background_resize (GPdfView *gpdf_view,
 }
 
 static void
-setup_page_transform (gdouble *transform, 
+setup_page_transform (gdouble *transform,
 		      gdouble x1, gdouble y1, gdouble x2, gdouble y2,
 		      gint angle)
 {
@@ -150,7 +151,7 @@ setup_page_transform (gdouble *transform,
 
 	art_affine_translate (translate, -x1, -y1);
 	art_affine_multiply (transform, translate, transform);
-}	
+}
 
 static gint
 canonical_multiple_of_90 (gint n)
@@ -168,20 +169,20 @@ canonical_multiple_of_90 (gint n)
 static void
 gpdf_free_history_node (gpointer node, gpointer user_data)
 {
-	g_free (node); 
+	g_free (node);
 }
 
-static void 
+static void
 gpdf_view_history_stack_page (GPdfView *gpdf_view)
 {
 	GPdfViewPrivate *priv = gpdf_view->priv;
-	GPdfViewHistory *hnode; 
+	GPdfViewHistory *hnode;
 
 	/* First delete forward history if one exists */
 	if (priv->forward_history) {
 		g_list_foreach (priv->forward_history,
 				gpdf_free_history_node,
-				NULL); 
+				NULL);
 		g_list_free (priv->forward_history);
 		priv->forward_history = (G_List *)NULL;
 	}
@@ -190,7 +191,33 @@ gpdf_view_history_stack_page (GPdfView *gpdf_view)
 	hnode = (GPdfViewHistory *)g_malloc (sizeof(GPdfViewHistory));
 	hnode->zoom = priv->zoom;
 	hnode->page = priv->current_page;
-	priv->backward_history = g_list_prepend (priv->backward_history, (gpointer)hnode); 
+	priv->backward_history = g_list_prepend (priv->backward_history, (gpointer)hnode);
+}
+
+void
+gpdf_view_clear_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	priv = gpdf_view->priv;
+
+	if (priv->forward_history) {
+		g_list_foreach (priv->forward_history,
+				gpdf_free_history_node,
+				NULL);
+		g_list_free (priv->forward_history);
+		priv->forward_history = NULL;
+	}
+
+	if (priv->backward_history) {
+		g_list_foreach (priv->backward_history,
+				gpdf_free_history_node,
+				NULL);
+		g_list_free (priv->backward_history);
+		priv->backward_history = NULL;
+	}	
 }
 
 static void
@@ -228,9 +255,11 @@ gpdf_view_link_action_goto (GPdfView *gpdf_view, LinkGoTo *link_goto)
 
 		delete dest;
 
-		if (page != gpdf_view->priv->requested_page)
+		if (page != gpdf_view->priv->requested_page) {
 			gpdf_view_goto_page (gpdf_view, page);
-	}	
+			gpdf_view_scroll_to_top (gpdf_view);
+		}
+	}
 }
 
 static void
@@ -242,7 +271,7 @@ gpdf_view_link_action_named (GPdfView *gpdf_view, LinkNamed *link_named)
 	g_return_if_fail (dynamic_cast <LinkNamed *> (link_named) != NULL);
 
 	action_name = link_named->getName ()->getCString ();
-	
+
 	if (strcmp (action_name, "NextPage") == 0) {
 		gpdf_view_page_next (gpdf_view);
 	} else if (strcmp (action_name, "PrevPage") == 0) {
@@ -368,7 +397,7 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 		gnome_canvas_root (GNOME_CANVAS (gpdf_view)),
 		GPDF_TYPE_LINKS_CANVAS_LAYER,
 		"links",
-		new Links (pdf_page->getAnnots (&obj), 
+		new Links (pdf_page->getAnnots (&obj),
 			   pdf_doc->getCatalog ()->getBaseURI ()),
 		NULL));
 	g_object_ref (G_OBJECT (priv->links_layer));
@@ -412,16 +441,41 @@ gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
 {
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
-	gpdf_view_history_stack_page (gpdf_view); 
+	gpdf_view_history_stack_page (gpdf_view);
 	gpdf_view_goto_page_no_history (gpdf_view, page);
 }
 
+void
+gpdf_view_scroll_to_top (GPdfView *gpdf_view)
+{
+	gint x, y;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view), &x, &y);
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), x, 0);
+}
+
+void
+gpdf_view_scroll_to_bottom (GPdfView *gpdf_view)
+{
+	gint x, y;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view), &x, &y);
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), x,
+				(GTK_LAYOUT (gpdf_view)->height
+				 - GTK_WIDGET (gpdf_view)->allocation.height));
+}
+
 void
 gpdf_view_page_prev (GPdfView *gpdf_view)
 {
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
 	gpdf_view_goto_page (gpdf_view, gpdf_view->priv->requested_page - 1);
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
@@ -430,6 +484,7 @@ gpdf_view_page_next (GPdfView *gpdf_view)
 	g_return_if_fail(GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
 	gpdf_view_goto_page (gpdf_view, gpdf_view->priv->requested_page + 1);
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
@@ -438,6 +493,7 @@ gpdf_view_page_first (GPdfView *gpdf_view)
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
 	gpdf_view_goto_page (gpdf_view, 1);
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
@@ -447,13 +503,14 @@ gpdf_view_page_last (GPdfView *gpdf_view)
 
 	gpdf_view_goto_page (gpdf_view,
 			     gpdf_view->priv->pdf_doc->getNumPages ());
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 gboolean
 gpdf_view_is_first_page (GPdfView *gpdf_view)
 {
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), FALSE);
-	
+
 	return gpdf_view->priv->requested_page == 1;
 }
 
@@ -474,60 +531,62 @@ gpdf_view_back_history (GPdfView *gpdf_view)
 {
 	GPdfViewPrivate *priv;
 	GPdfViewHistory *hnode;
-	int target_page = 0; 
-	double target_zoom = (double)0.0; 
-	
+	int target_page = 0;
+	double target_zoom = (double)0.0;
+
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 	priv = gpdf_view->priv;
 
 	g_return_if_fail (priv->backward_history != (G_List *)NULL);
-	g_return_if_fail (priv->backward_history->data != (gpointer)NULL); 
+	g_return_if_fail (priv->backward_history->data != (gpointer)NULL);
 
 	hnode = (GPdfViewHistory *)priv->backward_history->data;
-	target_page = hnode->page; 
-	target_zoom = hnode->zoom; 
+	target_page = hnode->page;
+	target_zoom = hnode->zoom;
 	priv->backward_history = g_list_remove (priv->backward_history,
 						(gpointer)hnode);
 	if (!g_list_length (priv->backward_history)) {
 		g_list_free (priv->backward_history);
 		priv->backward_history = (G_List *)NULL;
 	}
-	hnode->page = priv->current_page; 
-	hnode->zoom = priv->zoom; 
-	priv->forward_history = g_list_prepend (priv->forward_history, hnode); 
+	hnode->page = priv->current_page;
+	hnode->zoom = priv->zoom;
+	priv->forward_history = g_list_prepend (priv->forward_history, hnode);
 
 	gpdf_view_goto_page_no_history (gpdf_view, target_page);
-	/* TODO: change zoom level if different */
+	/* TODO: change zoom level and scroll offset if different */
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
 gpdf_view_forward_history (GPdfView *gpdf_view)
 {
 	GPdfViewPrivate *priv;
-	GPdfViewHistory *hnode; 
-	int target_page = 0; 
-	double target_zoom = (double)0.0; 
+	GPdfViewHistory *hnode;
+	int target_page = 0;
+	double target_zoom = (double)0.0;
 
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 	priv = gpdf_view->priv;
 
-	g_return_if_fail (priv->forward_history != (G_List *)NULL); 
-	g_return_if_fail (priv->forward_history->data != (gpointer)NULL); 
+	g_return_if_fail (priv->forward_history != (G_List *)NULL);
+	g_return_if_fail (priv->forward_history->data != (gpointer)NULL);
 	hnode = (GPdfViewHistory *)priv->forward_history->data;
-	target_page = hnode->page; 
-	target_zoom = hnode->zoom; 
+	target_page = hnode->page;
+	target_zoom = hnode->zoom;
 	priv->forward_history = g_list_remove (priv->forward_history,
 					       (gpointer)hnode);
 	if (!g_list_length (priv->forward_history)) {
 		g_list_free (priv->forward_history);
 		priv->forward_history = (G_List *)NULL;
 	}
-	hnode->page = priv->current_page; 
-	hnode->zoom = priv->zoom; 
-	priv->backward_history = g_list_prepend (priv->backward_history, hnode); 
+	hnode->page = priv->current_page;
+	hnode->zoom = priv->zoom;
+	priv->backward_history = g_list_prepend (priv->backward_history, hnode);
 
 	gpdf_view_goto_page_no_history (gpdf_view, target_page);
-	/* TODO: change zoom level if different */
+	/* TODO: change zoom level and scroll offset if different */
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 gboolean
@@ -538,7 +597,7 @@ gpdf_view_is_first_history(GPdfView *gpdf_view)
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), TRUE);
 	priv = gpdf_view->priv;
 
-	return (priv->backward_history == (G_List *)NULL); 
+	return (priv->backward_history == (G_List *)NULL);
 }
 
 gboolean
@@ -549,13 +608,13 @@ gpdf_view_is_last_history (GPdfView *gpdf_view)
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), TRUE);
 	priv = gpdf_view->priv;
 
-	return (priv->forward_history == (G_List *)NULL); 
+	return (priv->forward_history == (G_List *)NULL);
 }
 
 static int
 gpdf_view_get_page_count (GgvDocument *document)
 {
- 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), -1);
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), -1);
 
 	return GPDF_VIEW (document)->priv->pdf_doc->getNumPages ();
 }
@@ -566,7 +625,7 @@ gpdf_view_get_page_names (GgvDocument *document)
 	int page_count;
 	int i;
 	char **result;
-	
+
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), NULL);
 
 	page_count = gpdf_view_get_page_count (document);
@@ -575,7 +634,7 @@ gpdf_view_get_page_names (GgvDocument *document)
 	for (i = 0; i < page_count; ++i)
 		result [i] = g_strdup_printf ("%d", i+1);
 
-	return result;	
+	return result;
 }
 
 
@@ -731,7 +790,6 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Up:
 	case GDK_Up:
-		do_scroll = TRUE;
 		if (big_step) {
 			goto prev_screen;
 		} else {
@@ -741,7 +799,6 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Down:
 	case GDK_Down:
-		do_scroll = TRUE;
 		if (big_step) {
 			goto next_screen;
 		} else {
@@ -758,36 +815,38 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 	case GDK_KP_Delete:
 	case GDK_BackSpace:
 	prev_screen:
-		do_scroll = TRUE;
 		if (y <= 0) {
-			if (gpdf_view->priv->current_page > 1) {
-				gpdf_view_page_prev (gpdf_view);
-				/* 
-				 * FIXME height might be outdated for
-				 * new page
-				 */
-				y = GTK_LAYOUT (widget)->height - height;
-			}
+			if (gpdf_view->priv->current_page <= 1)
+				break;
+
+			gpdf_view_goto_page (
+				gpdf_view,
+				gpdf_view->priv->requested_page - 1);
+			gpdf_view_scroll_to_bottom (gpdf_view);
 		} else {
+			do_scroll = TRUE;
 			y -= height;
 		}
 		break;
 	case GDK_KP_Page_Down:
 	case GDK_Page_Down:
-		if (gpdf_view->priv->current_page < 
+		if (gpdf_view->priv->current_page <
 		    gpdf_view->priv->pdf_doc->getNumPages ())
 			gpdf_view_page_next (gpdf_view);
 		break;
 	case GDK_space:
 	next_screen:
-		do_scroll = TRUE;
 		if (y >= (gint)(GTK_LAYOUT (widget)->height) - height) {
-			if (gpdf_view->priv->current_page < 
-			    gpdf_view->priv->pdf_doc->getNumPages ()) {
-				gpdf_view_page_next (gpdf_view);
-				y = 0;
-			}
+			if (gpdf_view->priv->current_page >=
+			    gpdf_view->priv->pdf_doc->getNumPages ())
+				break;
+			
+			gpdf_view_goto_page (
+				gpdf_view,
+				gpdf_view->priv->requested_page + 1);
+			gpdf_view_scroll_to_top (gpdf_view);
 		} else {
+			do_scroll = TRUE;
 			y += height;
 		}
 		break;
@@ -838,7 +897,7 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 		cursor = gdk_cursor_new (GDK_FLEUR);
 		gdk_pointer_grab (widget->window, FALSE,
 				  (GdkEventMask)
-				  (GDK_POINTER_MOTION_MASK | 
+				  (GDK_POINTER_MOTION_MASK |
 				   GDK_POINTER_MOTION_HINT_MASK |
 				   GDK_BUTTON_RELEASE_MASK),
 				  NULL, cursor, event->time);
@@ -857,13 +916,13 @@ gpdf_view_button_release_event (GtkWidget *widget, GdkEventButton *event)
 	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
 
 	priv = GPDF_VIEW (widget)->priv;
-	
+
 	if (!priv->dragging || event->button != 1)
 		return FALSE;
 
 	priv->dragging = FALSE;
 	gdk_pointer_ungrab (event->time);
-	
+
 	return TRUE;
 }
 
@@ -889,7 +948,7 @@ gpdf_view_handle_drag_motion (GPdfView *gpdf_view, GdkEventMotion *event)
 	dx = priv->drag_anchor_x - x;
 	dy = priv->drag_anchor_y - y;
 
-	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), 
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view),
 				priv->drag_offset_x + dx,
 				priv->drag_offset_y + dy);
 
@@ -905,7 +964,7 @@ gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
 {
 
 	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
-	
+
 	if (GPDF_VIEW (widget)->priv->dragging)
 		gpdf_view_handle_drag_motion (GPDF_VIEW (widget), event);
 
@@ -965,6 +1024,7 @@ gpdf_view_set_pdf_doc (GPdfView *gpdf_view, PDFDoc *pdf_doc)
 
 	priv = gpdf_view->priv;
 
+	gpdf_view_clear_history (gpdf_view);
 	/*
 	 * All PDFDocs in GPdf code are owned by a GPdfPersistStream,
 	 * so don't delete our old pdf_doc
@@ -974,10 +1034,28 @@ gpdf_view_set_pdf_doc (GPdfView *gpdf_view, PDFDoc *pdf_doc)
 	priv->current_page = 0;
 	priv->output_dev->startDoc (pdf_doc->getXRef());
 	gpdf_view_page_background_show (gpdf_view);
-	/* FIXME clear history */
 	gpdf_view_goto_page_no_history (gpdf_view, 1);
 }
 
+/* Expects pathname in on-disk encoding (like Xpdf's saveAs) */
+void
+gpdf_view_save_as (GPdfView *gpdf_view, const gchar *pathname)
+{
+	GPdfViewPrivate *priv; 
+	GPOutputDev *gpOut;
+	GString *filename = new GString(pathname); 
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	priv = gpdf_view->priv;
+
+/*	gpdf_control_private_set_wait_cursor (priv->parent); */
+	
+	priv->pdf_doc->saveAs (filename);
+	
+/*	gpdf_control_private_reset_cursor (priv->parent); */
+}
+
 
 static void
 gpdf_view_dispose (GObject *object)
@@ -987,12 +1065,12 @@ gpdf_view_dispose (GObject *object)
 	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (object));
 
 	priv = GPDF_VIEW (object)->priv;
-	
+
 	if (priv->output_dev) {
 		delete priv->output_dev;
 		priv->output_dev = NULL;
 	}
-	
+
 	if (priv->print_context) {
 		g_object_unref (priv->print_context);
 		priv->print_context = NULL;
@@ -1017,7 +1095,7 @@ gpdf_view_finalize (GObject *object)
 	GPdfView *gpdf_view;
 
 	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (object));
-	
+
 	gpdf_view = GPDF_VIEW (object);
 
 	if (gpdf_view->priv) {
@@ -1092,6 +1170,16 @@ gpdf_view_instance_init (GPdfView *gpdf_view)
 	gpdf_view->priv = g_new0 (GPdfViewPrivate, 1);
 }
 
+static void
+gpdf_view_setup_at (GPdfView *gpdf_view)
+{
+	AtkObject *accessible;
+
+	accessible = gtk_widget_get_accessible (GTK_WIDGET (gpdf_view));
+	atk_object_set_role (accessible, ATK_ROLE_DRAWING_AREA);
+	atk_object_set_name (accessible, _("Document View"));
+}
+
 static void
 gpdf_view_setup_gesture_handler (GPdfView *gpdf_view)
 {
@@ -1169,6 +1257,7 @@ gpdf_view_construct (GPdfView *gpdf_view)
 
 	gpdf_view_setup_page_background (gpdf_view);
 	gpdf_view_setup_gesture_handler (gpdf_view);
+	gpdf_view_setup_at (gpdf_view);
 
 	priv = gpdf_view->priv;
 	priv->output_dev = new GPOutputDev ();
@@ -1207,7 +1296,7 @@ gpdf_view_get_type (void)
 		    gpdf_view_class_init_trampoline,
 		    NULL,		/* class_finalize */
 		    NULL,               /* class_data */
- 		    sizeof (GPdfView),
+		    sizeof (GPdfView),
 		    0,                  /* n_preallocs */
 		    (GInstanceInitFunc) gpdf_view_instance_init
 		};
