@@ -189,6 +189,18 @@ gpdf_page_control_setup_tooltips (GPdfPageControl *control)
 				"To jump to another page, enter its number."));
 }
 
+static void
+gpdf_page_control_setup_at (GPdfPageControl *control)
+{
+	AtkObject *accessible;
+
+	accessible = gtk_widget_get_accessible (GTK_WIDGET (control));
+	atk_object_set_name (accessible, _("Page"));
+
+	accessible = gtk_widget_get_accessible (control->priv->entry);
+	atk_object_set_name (accessible, _("Current page"));
+}
+
 void
 gpdf_page_control_instance_init (GPdfPageControl *control)
 {
@@ -197,4 +209,5 @@ gpdf_page_control_instance_init (GPdfPageControl *control)
 
 	gpdf_page_control_setup_widgets (control);
 	gpdf_page_control_setup_tooltips (control);
+	gpdf_page_control_setup_at (control);
 }
